﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Runtime.InteropServices;

using System.Threading;

namespace RenderLab2RenderBaby
{
	public unsafe partial class Form1 : Form
	{
		// Render surface
		Surface surface;

		// Time vars for keeping track of times
		public static DateTime pluginStartTime = DateTime.Now;
		public static DateTime pluginFinishTime = DateTime.Now;

		public static DateTime renderStartTime = DateTime.Now;
		public static DateTime renderFinishTime = DateTime.Now;

		DateTime lastTime = DateTime.Now;

		// Requested screen resolution
		int ScreenWidth, ScreenHeight;
		
		// Input state
		MouseState mouseState;
		KeyboardState keyboardState;

		// circular stacks to average times so they're not so flickering
		public static CircularStack pluginAverageTimes = new CircularStack(64);
		public static CircularStack renderAverageTimes = new CircularStack(64);

		// Import for input
		[DllImport("RenderBaby.dll", CharSet = CharSet.Unicode)]
		public static extern void UpdateInput(void* mouseState, void* keyboardState);

		/// <summary>
		/// Import for Init
		/// </summary>
		/// <param name="heightWidth"></param>
		[DllImport("RenderBaby.dll", CharSet = CharSet.Unicode)]
		public static extern void Init(int* heightWidth);

		public Form1()
		{
			mouseState = new MouseState();
			keyboardState = new KeyboardState();

			InitializeComponent();
		}


		/// <summary>
		/// Function to round doubles so they're not displaying so many dec places
		/// </summary>
		/// <param name="d">The double to round</param>
		/// <param name="dec">decimal places to round to</param>
		/// <returns></returns>
		double Round(double d, int dec)
		{
			int i = (int)((d * Math.Pow(10, dec)) + 0.5);
			return (double)i / Math.Pow(10, dec);
		}

		/// <summary>
		/// Update the info panel and call redraw
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void Timer1_Tick(object sender, EventArgs e)
		{
			label1.Text = "Plugin Time: " +
				Round(pluginAverageTimes.Average, 2) + "ms";
			txtRenderSpeed.Text = "Render Time: " +
				Round(renderAverageTimes.Average, 2) + "ms";

			txt16Bit.Text = "16-Bit: " + surface.dithering;
			txtVCL.Text = "VCL: " + surface.vcl;

			surface.Redraw();
		}

		/// <summary>
		/// Call reset when the picture box changes size
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void PictureBox1_Resize(object sender, EventArgs e)
		{
			if(surface != null)
				surface.Reset();
		}

		/// <summary>
		/// Call init and get the requested resolution
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void Form1_Load(object sender, EventArgs e)
		{
			int[] widthHeight = new int[2];

			fixed (int* hw = widthHeight)
			{
				Init(hw);
			}

			surface = new Surface(pictureBox1, widthHeight[0], widthHeight[1]);

			timer1.Start();
		}

		private void PictureBox1_Click(object sender, EventArgs e)
		{

		}

		/// <summary>
		/// Compute the mouse position when it moves
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void PictureBox1_MouseMove(object sender, MouseEventArgs e)
		{
			ComputeMousePos(e.X, e.Y);

			txtMouseInfo.Text = "Mouse: (" + 
				mouseState.ScreenX + ", " + mouseState.ScreenY + ")  (" +
				mouseState.PixelX + ", " + mouseState.PixelY + ")";

			UpdateMouseAndKeyboardState();
		}

		/// <summary>
		/// Update input
		/// </summary>
		void UpdateMouseAndKeyboardState()
		{
			fixed (int* mousePtr = &mouseState.Data[0], keyboardPtr = &keyboardState.Data[0])
			{
				UpdateInput(mousePtr, keyboardPtr);
			}
		}

		private void PictureBox1_MouseDown(object sender, MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Left)
				mouseState.ButtonDown(0);
			else if (e.Button == MouseButtons.Middle)
				mouseState.ButtonDown(1);
			else if (e.Button == MouseButtons.Right)
				mouseState.ButtonDown(2);

			UpdateMouseAndKeyboardState();
		}

		private void PictureBox1_MouseUp(object sender, MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Left)
				mouseState.ButtonUp(0);
			else if (e.Button == MouseButtons.Middle)
				mouseState.ButtonUp(1);
			else if (e.Button == MouseButtons.Right)
				mouseState.ButtonUp(2);

			UpdateMouseAndKeyboardState();
		}

		private void Form1_KeyDown(object sender, KeyEventArgs e)
		{
			if (e.KeyCode == Keys.Escape || e.KeyCode == Keys.Q)
			{
				timer1.Stop();
				surface.Dispose();
				this.Close();
			}
			else
			{
				// Change 16 bit dithering or integer
				if (e.KeyCode == Keys.F2)
					surface.ToggleDithering();
				// Change VCL vs C++
				else if (e.KeyCode == Keys.F3)
					surface.ToggleVCL();
				// Show/hide the info panel
				else if (e.KeyCode == Keys.N)
					groupBox1.Visible = !groupBox1.Visible;

				keyboardState.KeyDown(e.KeyCode);

				UpdateMouseAndKeyboardState();
			}
		}

		private void Form1_KeyUp(object sender, KeyEventArgs e)
		{
			keyboardState.KeyUp(e.KeyCode);

			UpdateMouseAndKeyboardState();
		}

		void ComputeMousePos(int x, int y)
		{
			ScreenWidth = pictureBox1.Width;
			ScreenHeight = pictureBox1.Height;

			mouseState.ScreenX = x;
			mouseState.ScreenY = y;

			float ratioX = (float)mouseState.ScreenX / (float)ScreenWidth;
			float ratioY = (float)mouseState.ScreenY / (float)ScreenHeight;

			mouseState.PixelX = (int)(surface.Width * ratioX);
			mouseState.PixelY = (int)(surface.Height * ratioY);
		}
	}
}
